/*
    Copyright (C) 2007 by Michael Dale Long
  
    This file is part of DuSTED.
  
    DuSTED is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
  
    DuSTED is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with DuSTED.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ARRAY_H
#define ARRAY_H

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <nds/jtypes.h>
#include "log.h"

#define DA_INITIAL_SIZE 20
#define DA_INCREMENT 10

template <class T, uint32 size> class FixedArray {
private:
	T mPool[size];
	uint32 mPoolTop;

public:
	FixedArray(void) {
		// Initialize variables
		mPoolTop = 0;
	}


	~FixedArray(void) {
		RemoveAll();
	}


	// Append an item to the list
	void Append(const T &val) {
		if (mPoolTop >= size) {
			LogFatal("FixedArray::Append() exceeded pool size.\n");
		}

		// Append to top of the array
		mPool[mPoolTop++] = val;
	}


	// Insert an item into the array
	void Insert(const T &val, uint32 pos) {
		if (mPoolTop >= size) {
			LogFatal("FixedArray::Insert() exceeded pool size.\n");
		}

		if (pos > mPoolTop) {
			LogFatal("FixedArray::Insert() exceeded array size.\n:");
		}

		// Handle the odd case
		if (pos == mPoolTop) {
			Append(val);
			return;
		}
	
		// Move the current variables out of the way safely
		memmove(&mPool[pos + 1], &mPool[pos], sizeof(T) * (mPoolTop - pos));
		mPool[pos] = val;
		mPoolTop++;
	}

	
	// Return an uninitialized element at the specified index (end of the array if -1)
	T &Alloc(int pos = -1) {
		if (mPoolTop >= size) {
			LogFatal("FixedArray::Append() exceeded pool size.\n");
		}

		if (pos == -1 || pos == mPoolTop) {
			// Allocate at the end of the array
			return mPool[++mPoolTop];
		} else {
			// Move the current variables out of the way safely
			memmove(&mPool[pos + 1], &mPool[pos], sizeof(T) * (mPoolTop - pos));
			mPoolTop++;
			return mPool[pos];
		}
	}

  
	// Remove a given item from the array
	void Remove(uint32 pos) {
		// Check bounds
		if (pos >= mPoolTop) {
			LogFatal("FixedArray::Remove():  Index is out of range.\n");
		}

		// Easy case
		if (pos == mPoolTop - 1) {
			mPoolTop--;
			return;
		}

		// Move the variables above the requested one down
		memmove(&mPool[pos], &mPool[pos + 1], sizeof(T) * (mPoolTop - (pos + 1)));
		mPoolTop--;
	}


	// Convenience function to remove a given item from the array
	void Remove(const T &val) {
		for (uint32 i = 0; i < mPoolTop; i++) {
			if (mPool[i] == val) {
				Remove(i);
				return;
			}
		}
		//TODO should we give a fatal error if it wasn't found?
	}

	// Resets the memory pool
	void RemoveAll(void) {
		mPoolTop = 0;
	}


	// Get the item at the given position
	T &Get(uint32 pos) {
		// Check bounds
		if (pos >= mPoolTop) {
			LogFatal("FixedArray::Get():  Exceed pool size.\n");
		}
		return mPool[pos];
	}


	// Copy a C array into the array
	void SetArray(uint32 count, T *elements) {
		if (count >= size) {
			LogFatal("FixedArray::SetArray():  Exceeded pool size.\n");
		}
		// Copy the data
		memcpy(mPool, elements, count * sizeof(T));
		mPoolTop = count;
	}


	// Shorthand way of calling Get()
	inline T &operator[](uint32 pos) { return Get(pos); }

	// Retrieve the size of the array
	inline uint32 Size(void) const { return mPoolTop; }

	// Retrieve a pointer to the array
	inline T *GetArrayPtr(void) { return mPool; }

	// Push a value onto the dynarray
	inline void Push(const T &val) { Insert(val, 0); }

	// Pop a value off the dynarray
	inline T Pop(void) { T tmp = Get(0); Remove(0); return tmp; }

	// Add an value only if it hasn't been added before
	bool AppendOnce(const T &val) {
		// Search the array for the value
		for (uint32 i = 0; i < mPoolTop; i++) {
			if (mPool[i] == val) {
				return false;
			}
		}
		
		if (mPoolTop >= size) {
			LogFatal("FixedArray::AppendOnce() exceeded pool size.\n");
		}

		// Append to top of the array
		mPool[mPoolTop++] = val;
		
		return true;
	}

};


template <class T> class DynArray {
private:
	T *mPool;
	uint32 mPoolSize;
	uint32 mPoolTop;

	void Resize(void) {
		if (mPool) {
			// Attempt to resize the pool
			ReserveSpace(mPoolSize + DA_INCREMENT);
		} else {
			// Reserve the initial size of the pool
			ReserveSpace();
		}
	}

public:
	DynArray(void) {
		// Initialize variables
		mPool = NULL;
		mPoolSize = 0;
		mPoolTop = 0;
	}


	~DynArray(void) {
		// Free pool memory
		if (mPool) {
			free(mPool);
			mPool = NULL;
		}
		mPoolSize = 0;
		mPoolTop = 0;
	}


	// Append an item to the list
	void Append(const T &val) {
		assert(mPoolTop <= mPoolSize);
		if (mPoolTop == mPoolSize) {
			Resize();
		}

		// Append to top of the array
		mPool[mPoolTop++] = val;
	}


	// Insert an item into the array
	void Insert(const T &val, uint32 pos) {
		assert(mPoolTop <= mPoolSize);
		if (mPoolTop == mPoolSize) {
			Resize();
		}

		// Handle the odd case
		if (pos == mPoolTop - 1) {
			Append(val);
			return;
		}

		// Move the current variables out of the way safely
		memmove(&mPool[pos + 1], &mPool[pos], sizeof(T) * (mPoolTop - pos));
		mPool[pos] = val;
		mPoolTop++;
	}

	
	// Remove a given item from the array
	void Remove(uint32 pos) {
		// Check bounds
		if (pos >= mPoolTop) {
			LogFatal("DynArray<T>::Remove(pos = %u):  pos is out of range.\n");
		}

		// Easy case
		if (pos == mPoolTop - 1) {
			mPoolTop--;
			return;
		}

		// Move the variables above the requested one down
		memmove(&mPool[pos], &mPool[pos + 1], sizeof(T) * (mPoolTop - (pos + 1)));
		mPoolTop--;
	}


	// Convenience function to remove a given item from the array
	void Remove(const T &val) {
		for (uint32 i = 0; i < mPoolSize; i++) {
			if (mPool[i] == val) {
				Remove(i);
				return;
			}
		}
		//TODO should we give a fatal error if it wasn't found?
	}

	// Resets the memory pool.  Free the memory pool if release is tree
	void RemoveAll(bool release = true) {
		mPoolTop = 0;
		if (release && mPool) {
			free(mPool);
			mPool = NULL;
			mPoolSize = 0;
		}
	}


	// Get the item at the given position
	T &Get(uint32 pos) {
		// Check bounds
		if (pos >= mPoolTop) {
			LogFatal("DynArray::Get():  Exceed pool size.\n");
		}
		return mPool[pos];
	}


	// Copy a C array into the array
	void SetArray(uint32 count, T *elements) {
		// Reserve enough space to hold the array
		ReserveSpace(count);
		
		// Copy the data
		memcpy(mPool, elements, count * sizeof(T));
		mPoolTop = count;
	}


	// Shorthand way of calling Get()
	inline T &operator[](uint32 pos) { return Get(pos); }


	// Retrieve the size of the array
	inline uint32 Size(void) const { return mPoolTop; }


	// Retrieve a pointer to the array
	inline T *GetArrayPtr(void) { return mPool; }


	// Reserves the given number of elements.  Never reduces the size of the pool.
	void ReserveSpace(uint32 size = DA_INITIAL_SIZE) {
		// Calculate the new size of the pool
		uint32 newsize = ((uint32) ceilf((float) size / (float) DA_INCREMENT)) * DA_INCREMENT;
		if (mPoolSize > newsize) {
			// Don't do anything
			return;
		}
		// Attempt to resize the pool
		T *ptr;
		if ((ptr = (T *) realloc(mPool, sizeof(T) * newsize))) {
			mPoolSize = newsize;
			mPool = ptr;
		} else {
			LogFatal("DynArray<T>::Resize(%u) failed.\n", size);
		}
	}


	// Push a value onto the dynarray
	inline void Push(const T &val) { Insert(val, 0); }

	// Pop a value off the dynarray
	inline T Pop(void) { T tmp = Get(0); Remove(0); return tmp; }

	// Add an item only if it hasn't been added before
	bool AppendOnce(const T &val) {
		// Search the array for the val
		for (int i = 0; i < mPoolTop; i++) {
			if (mPool[i] == val) {
				// Found it, not adding again
				return false;
			}
		}
		
		// Append to top of the array
		Append(val);
		
		return true;
	}
};

#endif // ! ARRAY_H

