/*
    Copyright (C) 2007 by Michael Dale Long
  
    This file is part of DuSTED.
  
    DuSTED is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
  
    DuSTED is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with DuSTED.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <nds.h>
#include <nds/arm9/console.h>
#include "dusted.h"
#include "engine.h"
#include "window.h"

// Global engine pointer
Engine *gEngine;

#ifdef DEBUG
static DynArray<void *> mPtrs;
#endif

#ifdef SYSTEMMODE_INTERRUPTS
extern "C" void IrqThunk(void) {
	//TODO check interrupt type.  Right now we assume it's a vblank
	gEngine->VBlankHandler();
}
#endif // SYSTEMMODE_INTERRUPTS


// IRQ Handler function.  Handles all IRQs.  Runs in IRQ mode (easy to overflow the stack on hw, but runs on emulators w/o problems).
extern "C" void IrqHandler(void) {	
  // Disable interrupts
  REG_IME = 0;

#ifdef SYSTEMMODE_INTERRUPTS
	asm volatile("     stmfd     sp!, {r0-r2}              ");     // Save regs on IRQ stack. 
	asm volatile("     mrs       r0, cpsr                  ");     // Grab CPSR which is in IRQ mode right now. 
	asm volatile("     orr       r1, r0, #0x1f             ");     // 0x1f = SYS mode. No need to clear first since all bits are set. 
	asm volatile("     msr       cpsr, r1                  ");     // Switch to SYS mode. 
	asm volatile("     stmfd     sp!, {r0-r3, r12, lr}     ");     // Save regs as per APCS, on user stack. 
	asm volatile("     bl        IrqThunk                  ");     // Call the IRQ Handler
	asm volatile("     ldmfd     sp!, {r0-r3, r12, lr}     ");     // Load regs back from user stack. 
	asm volatile("     msr       cpsr, r0                  ");     // Switch back to IRQ mode. 
	asm volatile("     ldmfd     sp!, {r0-r1}              ");     // Restore regs from IRQ stack. 
#else 
	//TODO check interrupt type.  Right now we assume it's a vblank
	gEngine->VBlankHandler();
#endif // SYSTEMMODE_INTERRUPTS

	// Tell the DS we handled the VBLANK interrupt
  REG_IF = REG_IF;
  VBLANK_INTR_WAIT_FLAGS = REG_IF | REG_IE;

  // Enable interrupts
  REG_IME = 1;
}


// Override new to trap failed allocations.  This way we never have to worry about checking for NULL from new.
// It also includes some debug code for trapping bad deletes.
void *operator new(size_t size) {
	void *ptr = malloc(size);
	if (ptr == NULL) {
		LogFatal("::new():  Failed to malloc %u bytes.\n", size);
	}
#ifdef DEBUG
	// Add new pointer to valid pointers list
	mPtrs.Append(ptr);
#endif

	return ptr;
}


// Counterpart to above new operator.  Also includes debug code for trapping bad deletes.
void operator delete(void *ptr) {
#ifdef DEBUG
	// NULL pointer check
	if (!ptr) {
		LogFatal("::delete():  Given NULL pointer.\n");
	}
	
	// Valid pointer check
	uint32 i;
	for (i = 0; i < mPtrs.Size(); i++) {
		if (mPtrs[i] == ptr) {
			break;
		}
	}
	if (i == mPtrs.Size()) {
		LogFatal("::delete():  Given invalid pointer.\n");
	}

	// Remove it from the valid pointers list
	mPtrs.Remove(i);
#endif // DEBUG

	free(ptr);
}


void Engine::Init(void) {
	powerON(POWER_ALL);

	LogInit();

	// Map the GBA Cart to the ARM9 processor
	REG_EXMEMCNT &= ~ARM7_OWNS_ROM;

	// Allocate PBuffer
	mPBuffer = (char *) malloc(PBUFFER_LEN);
#ifdef NDS_SIMULATOR
	// Simulator can always save
	//TODO maybe we could include the actual rom in memory for gbfs type stuff?
	//TODO maybe this could be a command line option for testing?
	mCartIsValid = true;
#else
	char buffer[5] = { 0 };
	mCartIsValid = false;
	// Read the game cart header
	memcpy8(buffer, (char *) (0x080000AC), 4);
	//TODO should check this on cart interrupt as well.  Live updating of the save buffer would be cool.
	// See if we're running with a GBA homebrew cart inserted
	if (!strcmp(buffer, "PASS")) {
		// We can save
		mCartIsValid = true;
	}
#endif

	// Set up the primary display
	videoSetMode(MODE_FB0);
	vramSetBankA(VRAM_A_LCD);

	// Set up the IRQs
	irqInit();
	//TODO re-enable IrqHandler
	//irqSet(IRQ_VBLANK, IrqHandler);
	irqEnable(IRQ_VBLANK);

	lcdSwap();

	ShowConsole(false);
	LogInfo("DuSTED v" VERSION " ready.\n");

	// Report savegame status
	if (mCartIsValid) {
		LogInfo("Load/save enabled.\n");
		// Load save data into RAM
		ReadPBuffer();
	} else {
		LogInfo("Load/save disabled.\n");
	}
}


void Engine::Loop(void) {
	while (1) {
		if (mFocusStack.Size() == 0) {
			LogFatal("Engine::Loop():  mFocusStack is empty.\n");
		}

		// Give the focus window a frame tick
		mFocusStack[0]->Tick();
		mFocusStack[0]->Render();  //TODO remove this when IrqHandler is fixed

		// Sleep until the beginning of the next frame
		swiWaitForVBlank();
	}
}


void Engine::VBlankHandler(void) {
	// Give the focus window a frame tick
	if (mFocusStack.Size() > 0) { //TODO I'd like to get rid of this, but the interrupt could occur while the stack is empty (between Init() and Loop()).
		mFocusStack[0]->Render();
	}
}


void Engine::memcpy8(char *dest, const char *src, int len) {
	// Byte by byte copy
	while (len--) {
		*dest++ = *src++;
	}
}


bool Engine::ReadPBuffer(int start, int len) {
	if (!mCartIsValid) {
		return false;
	}

	if (!IsSRAMFormatted()) {
		return false;
	}

	if (start < 0 || len <= 0 || (int) len + (int) start > PBUFFER_LEN) {
		LogFatal("ReadPBuffer:  Bad len(%u) and/or start(%u)\n", len, start);
	}

	memcpy8(mPBuffer + start, (char *) (SRAM + start), len);

	return true;
}


bool Engine::WritePBuffer(int start, int len) {
	if (!mCartIsValid) {
		return false;
	}

	if (!IsSRAMFormatted()) {
		return false;
	}

	if (start < 0 || len <= 0 || (int) len + (int) start > PBUFFER_LEN) {
		LogFatal("WritePBuffer:  Bad len(%u) and/or start(%u)\n", len, start);
	}
	
	memcpy8((char *) (SRAM + start), mPBuffer + start, len);

	return true;
}


bool Engine::IsSRAMFormatted(void) {
	if (!mCartIsValid) {
		return false;
	}

	char buffer[8];

	// Read in the header
	memcpy8(buffer, (const char *) SRAM, 8);

	// Check the signature
	if (strncmp(buffer, SRAM_SIGNATURE, 4)) {
		// Signature didn't match
		return false;
	}

	// Last four bytes are 
	uint32 *version = (uint32 *) &buffer[4];
	if (*version != SRAM_VERSION) {
		// Bad version
		return false;
	}

	return true;
}


bool Engine::FormatSRAM(void) {
	if (!mCartIsValid) {
		return false;
	}

	// Setup the signature and the version
	strncpy(mPBuffer, SRAM_SIGNATURE, 4);
	uint32 *version = (uint32 *) &mPBuffer[4];
	*version = SRAM_VERSION;

	// Clear the rest of memory
	memset(&mPBuffer[8], 0, PBUFFER_LEN - 8);

	// Write the data to the SRAM
	memcpy8((char *) SRAM, mPBuffer, PBUFFER_LEN);

	return true;
}

