/*
    Copyright (C) 2007 by Michael Dale Long
  
    This file is part of DuSTED.
  
    DuSTED is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
  
    DuSTED is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with DuSTED.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdarg.h>
#include <nds.h>
#include <nds/arm9/console.h>
#include "dusted.h"
#include "array.h"
#include "log.h"

#define CONSOLE_WIDTH 32
#define CONSOLE_HEIGHT 24

// Buffer to hold the console text when inactive
static char log_buffer[CONSOLE_HEIGHT][CONSOLE_WIDTH];
static int16 log_buffer_top;
static int16 log_buffer_bottom;
static char fatal_buffer[FATAL_BUFFER_SIZE];
static bool console_active;

void LogInfo(const char *format, ...) {
	int index = log_buffer_bottom++;
	if (log_buffer_bottom == CONSOLE_HEIGHT) {
		log_buffer_bottom = 0;
	}
	if (log_buffer_bottom == log_buffer_top) {
		log_buffer_top++;
		if (log_buffer_top == CONSOLE_HEIGHT) {
			log_buffer_top = 0;
		}
	}

	// Create the given error message
	va_list ap;
	va_start(ap, format);
	vsnprintf(log_buffer[index], CONSOLE_WIDTH - 1, format, ap);
	log_buffer[index][CONSOLE_WIDTH - 1] = 0; // Ensure null-terminated
	va_end(ap);

#ifdef NDS_SIMULATOR
	simWarning("%s", log_buffer[index]);
#else
	if (console_active) {
		iprintf("%s", log_buffer[index]);
	}
#endif
}


void LogFatal(const char *format, ...) {
	if (!console_active) {
		ShowConsole(false);
	}

	BG_PALETTE_SUB[0] = RGB15(0,0,31);

  // Create the given error message
	va_list ap;
	va_start(ap, format);
	vsnprintf(fatal_buffer, FATAL_BUFFER_SIZE - 1, format, ap);
	fatal_buffer[FATAL_BUFFER_SIZE - 1] = 0; // Ensure null-terminated
	va_end(ap);
	
	//TODO remove this once the console works in the simulator
#ifdef NDS_SIMULATOR
	simWarning("FATAL ERROR!\n\n");
	
	simWarning("%s", fatal_buffer);
#else
	iprintf("FATAL ERROR!\n\n");

	iprintf("%s", fatal_buffer);
#endif

	while(1) {
		//TODO: Do I need to put the ARM7 processor to sleep, too?
		swiWaitForVBlank();
	}
}


//TODO is this really how we want LogAssert implemented?
void LogAssert(bool test, const char *format, ...) {
	if (!test) {
		return;
	}

	if (!console_active) {
		ShowConsole(false);
	}

	BG_PALETTE_SUB[0] = RGB15(0,0,31);

	// Create the given error message
	va_list ap;
	va_start(ap, format);
	vsnprintf(fatal_buffer, FATAL_BUFFER_SIZE - 1, format, ap);
	fatal_buffer[FATAL_BUFFER_SIZE - 1] = 0; // Ensure null-terminated
	va_end(ap);
	
	//TODO remove this once the console works in the simulator
#ifdef NDS_SIMULATOR
	simWarning("FATAL ERROR!\n\n");
	
	simWarning("%s", fatal_buffer);
#else
	iprintf("FATAL ERROR!\n\n");

	iprintf("%s", fatal_buffer);
#endif

	while(1) {
		//TODO: Do I need to put the ARM7 processor to sleep, too?
		swiWaitForVBlank();
	}
}


void ShowConsole(bool draw_buffer) {
	// Setup the sub display, just in case
	videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
	vramSetBankC(VRAM_C_SUB_BG);
	SUB_BG0_CR = BG_MAP_BASE(31);
	// Clear the palette
	//TODO this causes memory corruption on the simulator.  Probably an allocation problem (sim-only)
	memset(BG_PALETTE_SUB, 0, 510);
	// Set the last entry to white
	BG_PALETTE_SUB[255] = RGB15(31,31,31);

	consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(31), (u16*)CHAR_BASE_BLOCK_SUB(0), 16);

	console_active = true;
		
#ifndef NDS_SIMULATOR // Don't bother with this until the console is simulated properly
	// Clear the console
	// /x1b[line;columnH
	iprintf("\x1b[2J");
	
	if (draw_buffer) {
		int i = log_buffer_top;
		while (i != log_buffer_bottom) {
			iprintf("%s", log_buffer[i]);
			i++;
			if (i > CONSOLE_HEIGHT) {
				i = 0;
			}
		}
	}
#endif
}


void HideConsole(void) {
	console_active = false;
}


// Initialize the log variables
void LogInit(void) {
	memset(log_buffer, 0, sizeof(char) * CONSOLE_WIDTH * CONSOLE_HEIGHT);
	log_buffer_top = 0;
	log_buffer_bottom = 0;
	memset(fatal_buffer, 0, FATAL_BUFFER_SIZE);
	console_active = false;
}

